from multiprocessing import Pool, cpu_count
import pickle
from time import time
from fw_ddsm.household import *
from fw_ddsm.tracker import *
from fw_ddsm.scripts import household_generation, household_scheduling


class Community:

    def __init__(self, num_intervals=no_intervals, num_periods=no_periods):
        self.num_intervals = num_intervals
        self.num_periods = num_periods
        self.num_intervals_periods = int(num_intervals / num_periods)
        self.num_households = 0
        self.households = dict()
        self.tracker = Tracker()
        self.final = Tracker()
        self.scheduling_method = ""
        self.preferred_demand_profile = []

    def read(self, scheduling_method, read_from_folder="data/",
             inconvenience_cost_weight=None, num_dependent_tasks=None, ensure_dependent=False,
             date_time=None):
        if not read_from_folder.endswith("/"):
            read_from_folder += "/"

        self.households = dict()
        self.households, self.preferred_demand_profile \
            = self.__existing_households(file_path=read_from_folder, date_time=date_time,
                                         inconvenience_cost_weight=inconvenience_cost_weight,
                                         num_dependent_tasks=num_dependent_tasks, ensure_dependent=ensure_dependent)
        if s_demand in self.households:
            self.num_households = len(self.households) - 1
        self.new_community_tracker(scheduling_method=scheduling_method)
        print("0. The community is read. ")

        return self.preferred_demand_profile

    def new(self, file_preferred_demand_profile, file_demand_list, scheduling_method,
            num_intervals=no_intervals, num_households=no_households,
            max_demand_multiplier=maximum_demand_multiplier,
            num_tasks_dependent=no_tasks_dependent, ensure_dependent=False,
            full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
            semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
            fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
            inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
            write_to_file_path=None, date_time=None):

        self.scheduling_method = scheduling_method
        self.num_intervals = num_intervals
        self.num_households = num_households

        households = dict()
        aggregate_demand_profile = [0] * num_intervals
        preferred_demand_profile = genfromtxt(file_preferred_demand_profile, delimiter=',', dtype="float")
        list_of_devices_power = genfromtxt(file_demand_list, delimiter=',', dtype="float")
        for h in range(num_households):
            household, household_demand_profile \
                = household_generation.new_household(num_intervals=num_intervals,
                                                     preferred_demand_profile=preferred_demand_profile,
                                                     list_of_devices_power=list_of_devices_power,
                                                     max_demand_multiplier=max_demand_multiplier,
                                                     num_tasks_dependent=num_tasks_dependent,
                                                     ensure_dependent=ensure_dependent,
                                                     full_flex_task_min=full_flex_task_min,
                                                     full_flex_task_max=full_flex_task_max,
                                                     semi_flex_task_min=semi_flex_task_min,
                                                     semi_flex_task_max=semi_flex_task_max,
                                                     fixed_task_min=fixed_task_min,
                                                     fixed_task_max=fixed_task_max,
                                                     inconvenience_cost_weight=inconvenience_cost_weight,
                                                     max_care_factor=max_care_factor,
                                                     household_id=h)
            aggregate_demand_profile = [x + y for x, y in zip(household_demand_profile, aggregate_demand_profile)]

            households[h] = household.copy()
            households[h][k_tracker] = Tracker()
            households[h][k_tracker].new()
            households[h][k_tracker].update(num_record=0, demands=household_demand_profile, penalty=0)

        self.households = households
        self.preferred_demand_profile = aggregate_demand_profile
        self.new_community_tracker(scheduling_method=scheduling_method)

        if write_to_file_path is not None:
            self.write_to_file(write_to_file_path, date_time=date_time)
        else:
            self.write_to_file("data/")
        print("0. The community is created. ")
        return aggregate_demand_profile

    def new_community_tracker(self, scheduling_method):
        self.tracker = Tracker()
        self.tracker.new(name=f"{scheduling_method}_community")
        self.tracker.update(num_record=0, penalty=0, run_time=0)
        self.final = Tracker()
        self.final.new(name=f"{scheduling_method}_community_final")
        self.tracker.update(num_record=0, penalty=0, run_time=0)

    def write_to_file(self, folder, date_time=None):

        if not folder.endswith("/"):
            folder += "/"
        folder += "data/"
        path = Path(folder)
        if not path.exists():
            path.mkdir(mode=0o777, parents=True, exist_ok=False)

        self.households[s_demand] = self.preferred_demand_profile
        if date_time is None:
            file_name = f"{folder}{file_community_pkl}"
        else:
            file_name = f"{folder}{date_time}_{file_community_pkl}"
        with open(file_name, 'wb+') as f:
            pickle.dump(self.households, f, pickle.HIGHEST_PROTOCOL)
        del self.households[s_demand]
        f.close()

    def schedule(self, num_iteration, prices, scheduling_method, model=None, solver=None, search=None, households=None,
                 num_cpus=None, timeout=time_out):

        prices = self.__convert_price(prices)

        if households is None:
            households = self.households
        results = self.__schedule_multiple_processing(households=households, prices=prices,
                                                      scheduling_method=scheduling_method,
                                                      model=model, solver=solver, search=search,
                                                      num_cpus=num_cpus, timeout=timeout)

        aggregate_demand_profile, weighted_total_inconvenience, time_scheduling_iteration \
            = self.__retrieve_scheduling_results(results=results, num_iteration=num_iteration)

        self.tracker.update(num_record=num_iteration, penalty=weighted_total_inconvenience,
                            run_time=time_scheduling_iteration)

        return aggregate_demand_profile, weighted_total_inconvenience, time_scheduling_iteration

    def finalise_schedule(self, scheduling_method, start_probability_distribution, num_sample=0):
        final_aggregate_demand_profile = [0] * self.num_intervals
        final_total_inconvenience = 0
        total_demand = 0
        for household in self.households.values():
            chosen_demand_profile, chosen_penalty \
                = Household.finalise_household(self=Household(), household_tracker_data=household[k_tracker].data,
                                               probability_distribution=start_probability_distribution)
            final_aggregate_demand_profile \
                = [x + y for x, y in zip(chosen_demand_profile, final_aggregate_demand_profile)]
            final_total_inconvenience += chosen_penalty
            total_demand += sum(chosen_demand_profile)

        self.final.update(num_record=num_sample, demands=final_aggregate_demand_profile,
                          penalty=final_total_inconvenience)

        return final_aggregate_demand_profile, final_total_inconvenience

    def __convert_price(self, prices):
        num_periods = len(prices)
        num_intervals_period = int(self.num_intervals / num_periods)
        if num_periods != self.num_intervals:
            prices = [p for p in prices for _ in range(num_intervals_period)]
        else:
            prices = [p for p in prices]

        return prices

    def __existing_households(self, file_path, date_time=None, inconvenience_cost_weight=None,
                              num_dependent_tasks=None, ensure_dependent=False):
        # ---------------------------------------------------------------------- #
        # ---------------------------------------------------------------------- #
        if date_time is None:
            file_name = f"{file_path}data/{file_community_pkl}"
        else:
            file_name = f"{file_path}data/{date_time}_{file_community_pkl}"
        with open(file_name, 'rb') as f:
            households = pickle.load(f)
        f.close()
        preferred_demand_profile = households.pop(s_demand)

        for key, household in households.items():
            household_tracker = Tracker()
            household_tracker.new()
            household_tracker.update(num_record=0, demands=household[s_demand], penalty=0)
            household[k_tracker] = household_tracker

            if inconvenience_cost_weight is not None:
                household[h_incon_weight] = inconvenience_cost_weight

            if num_dependent_tasks is not None:
                num_intervals = len(household[s_demand])
                durations = household[h_durs]
                num_total_tasks = len(durations)
                preferred_starts = household[h_psts]
                earliest_starts = household[h_ests]
                latest_ends = household[h_lfs]
                no_precedences, precedors, succ_delays \
                    = household_generation.new_dependent_tasks(
                    num_intervals=num_intervals, num_tasks_dependent=num_dependent_tasks,
                    ensure_dependent=ensure_dependent,
                    num_total_tasks=num_total_tasks,
                    preferred_starts=preferred_starts, durations=durations, earliest_starts=earliest_starts,
                    latest_ends=latest_ends)
                household[h_no_precs] = no_precedences
                household[h_precs] = precedors.copy()
                household[h_succ_delay] = succ_delays.copy()

            households[key] = household.copy()

        return households, preferred_demand_profile

    def __schedule_multiple_processing(self, households, prices, scheduling_method, model, solver, search,
                                       num_cpus=None, timeout=time_out):
        if num_cpus is not None:
            pool = Pool(num_cpus)
        else:
            pool = Pool()
        results = pool.starmap(Household.schedule_household,
                               [(Household(), prices, scheduling_method, household,
                                 self.num_intervals, model, solver, search, timeout)
                                for household in households.values()])
        # parameter order: prices, scheduling_method, household, num_intervals, model, solver, search
        pool.close()
        pool.join()
        return results

    def __retrieve_scheduling_results(self, results, num_iteration):
        aggregate_demand_profile = [0] * self.num_intervals
        total_weighted_inconvenience = 0
        time_scheduling_iteration = 0
        total_demand = 0
        for res in results:
            # print(res)
            key = res[h_key]
            demands_household = res[s_demand]
            weighted_penalty_household = res[s_penalty]
            time_household = res[t_time]
            total_demand += sum(demands_household)

            aggregate_demand_profile = [x + y for x, y in zip(demands_household, aggregate_demand_profile)]
            total_weighted_inconvenience += weighted_penalty_household
            time_scheduling_iteration += time_household

            # update each household's tracker
            self.households[key][k_tracker].update(num_record=num_iteration, demands=demands_household,
                                                   penalty=weighted_penalty_household)

        return aggregate_demand_profile, total_weighted_inconvenience, time_scheduling_iteration



from math import ceil
from time import time
from fw_ddsm.parameter import *
from fw_ddsm.scripts.custom_functions import *


def prices_and_cost(aggregate_demand_profile, pricing_table, cost_function=cost_function_type):
    prices = []
    consumption_cost = 0

    price_levels = pricing_table[p_price_levels]
    for demand_period, demand_level_period in \
            zip(aggregate_demand_profile, pricing_table[p_demand_table].values()):
        demand_level = list(demand_level_period.values())
        level = bisect_left(demand_level, demand_period)
        if level != len(demand_level):
            price = price_levels[level]
        else:
            price = price_levels[-1]
        prices.append(price)

        if "piece-wise" in cost_function and level > 0:
            consumption_cost += demand_level[0] * price_levels[0]
            consumption_cost += (demand_period - demand_level[level - 1]) * price
            consumption_cost += sum([(demand_level[i] - demand_level[i - 1]) *
                                     price_levels[i] for i in range(1, level)])
        else:
            consumption_cost += demand_period * price

    consumption_cost = round(consumption_cost, 2)

    return prices, consumption_cost


def find_step_size(num_iteration, pricing_method, pricing_table, aggregate_demand_profile, aggregate_inconvenience,
                   demand_profile_fw_pre, inconvenience_fw_pre, price_fw_pre, cost_fw_pre, min_step_size=min_step,
                   ignore_tiny_step=False, roundup_tiny_step=False):
    time_begin = time()

    price_fw = price_fw_pre[:]
    cost_fw = cost_fw_pre
    demand_profile_fw = demand_profile_fw_pre[:]
    inconvenience_fw = inconvenience_fw_pre
    change_of_inconvenience = aggregate_inconvenience - inconvenience_fw_pre
    demand_profile_changed = [d_n - d_p for d_n, d_p in zip(aggregate_demand_profile, demand_profile_fw_pre)]

    step_size_final = 0
    gradient = -999
    num_itrs = 0
    while gradient < 0 and step_size_final < 1:
        step_profile = []
        for dp, dn, demand_levels_period in \
                zip(demand_profile_fw_pre, aggregate_demand_profile, pricing_table[p_demand_table].values()):
            d_levels = list(demand_levels_period.values())[:-1]
            min_demand_level = min(d_levels)
            max_demand_level = d_levels[-1]
            second_max_demand_level = d_levels[-2]
            if dn < dp < min_demand_level or dp < dn < min_demand_level or dn > dp > second_max_demand_level \
                    or dp > dn > max_demand_level or dn == dp:
                step = 1
            else:
                dd = dn - dp
                dl = find_ge(d_levels, dp) + 0.01 if dd > 0 else find_le(d_levels, dp) - 0.01
                step = (dl - dp) / dd
                if ignore_tiny_step:
                    step = step if step > min_step_size else 1
                elif roundup_tiny_step:
                    step = ceil(step * 1000) / 1000
                step = max(step, min_step_size)
            step_profile.append(step)
        step_size_incr = min(step_profile)

        demand_profile_fw_temp = [d_p + (d_n - d_p) * step_size_incr for d_p, d_n in
                                  zip(demand_profile_fw_pre, aggregate_demand_profile)]
        price_fw_temp, cost_fw_temp = prices_and_cost(aggregate_demand_profile=demand_profile_fw_temp,
                                                      pricing_table=pricing_table,
                                                      cost_function=cost_function_type)
        gradient = change_of_inconvenience \
                   + sum([d_c * p_fw for d_c, p_fw in zip(demand_profile_changed, price_fw_temp)])

        demand_profile_fw_pre = demand_profile_fw_temp[:]
        step_size_final_temp = step_size_final + step_size_incr
        if gradient < 0 and step_size_final_temp < 1:
            step_size_final = step_size_final_temp
            demand_profile_fw = demand_profile_fw_temp[:]
            price_fw = price_fw_temp[:]
            cost_fw = cost_fw_temp
            inconvenience_fw = inconvenience_fw_pre + step_size_final * change_of_inconvenience
            num_itrs += 1

    print(f"{num_iteration}. "
          f"Best step size {round(step_size_final, 3)}, "
          f"{num_itrs} iterations, cost {cost_fw}, "
          f"using {pricing_method}")
    time_fw = time() - time_begin
    return demand_profile_fw, step_size_final, price_fw, cost_fw, inconvenience_fw, time_fw


def compute_start_time_probabilities(history_steps):
        prob_dist = []
        if history_steps[0] == 0 or history_steps[0] ==1:
            del history_steps[0]

        for alpha in history_steps:
            if not prob_dist:
                prob_dist.append(1 - alpha)
                prob_dist.append(alpha)
            else:
                prob_dist = [p_d * (1 - alpha) for p_d in prob_dist]
                prob_dist.append(alpha)

        return prob_dist



import pickle
from numpy.random import choice
from json import dumps, load
from pathlib import Path

from numpy import genfromtxt
from fw_ddsm.scripts import household_generation
from fw_ddsm.scripts import household_scheduling
from fw_ddsm.tracker import *


class Household:

    def __init__(self, num_intervals=no_intervals, num_periods=no_periods):
        self.num_intervals = num_intervals
        self.num_periods = num_periods
        self.num_intervals_periods = int(num_intervals / num_periods)
        self.scheduling_method = ""
        self.tasks = dict()
        self.household_id = 0
        self.household_tracker = Tracker()
        self.household_final = Tracker()

    def read_household(self, scheduling_method, read_from_folder="households", household_id=0):
        self.scheduling_method = scheduling_method

        if not read_from_folder.endswith("/"):
            read_from_folder += "/"
        with open(f"{read_from_folder}household{household_id}.txt", 'r') as f:
            household = load(f)
        f.close()
        self.tasks = household
        self.household_id = household_id

        self.household_tracker = Tracker()
        self.household_tracker.new(name=f"h{household_id}")
        self.household_tracker.update(num_record=0, demands=self.tasks[s_demand], penalty=0)
        self.household_final = Tracker()
        self.household_final.new(name=f"h{household_id}_final")

        print(f"0. Household{household[h_key]} is read.")
        return self.tasks, self.household_tracker

    def new(self, num_intervals, scheduling_method,
            preferred_demand_profile=None, list_of_devices_power=None,
            preferred_demand_profile_csv=None, list_of_devices_power_csv=None,
            max_demand_multiplier=maximum_demand_multiplier,
            num_tasks_dependent=no_tasks_dependent,
            full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
            semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
            fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
            inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
            write_to_folder=None, household_id=0):

        self.scheduling_method = scheduling_method
        self.household_id = 0

        if preferred_demand_profile is None and preferred_demand_profile_csv is None:
            print("Please provide a preferred demand profile or the csv. ")
        if list_of_devices_power is None and list_of_devices_power_csv is None:
            print("Please provide the power rates of the tasks. ")
        if preferred_demand_profile_csv is not None:
            preferred_demand_profile = genfromtxt(preferred_demand_profile_csv, delimiter=',', dtype="float")
        if list_of_devices_power_csv is not None:
            list_of_devices_power = genfromtxt(list_of_devices_power_csv, delimiter=',', dtype="float")

        tasks, household_demand_profile \
            = household_generation.new_household(num_intervals=num_intervals,
                                                 preferred_demand_profile=preferred_demand_profile,
                                                 list_of_devices_power=list_of_devices_power,
                                                 max_demand_multiplier=max_demand_multiplier,
                                                 num_tasks_dependent=num_tasks_dependent,
                                                 full_flex_task_min=full_flex_task_min,
                                                 full_flex_task_max=full_flex_task_max,
                                                 semi_flex_task_min=semi_flex_task_min,
                                                 semi_flex_task_max=semi_flex_task_max,
                                                 fixed_task_min=fixed_task_min,
                                                 fixed_task_max=fixed_task_max,
                                                 inconvenience_cost_weight=inconvenience_cost_weight,
                                                 max_care_factor=max_care_factor,
                                                 household_id=household_id)

        if write_to_folder is not None:
            self.save_to_file(tasks=tasks, folder=write_to_folder, household_id=household_id)
        self.tasks = tasks.copy()

        self.household_tracker = Tracker()
        self.household_tracker.new(name=f"h{household_id}")
        self.household_tracker.update(num_record=0, demands=household_demand_profile, penalty=0)
        self.household_final = Tracker()
        self.household_final.new(name=f"h{household_id}_final")

        # print(f"Household{household_id} is created.")
        return self.tasks, self.household_tracker

    def save_to_file(self, tasks, folder, household_id=0):
        if not folder.endswith("/"):
            folder += "/"
        file_name = f"h{household_id}.txt"

        path = Path(folder)
        if not Path(folder).exists():
            path.mkdir(mode=0o777, parents=True, exist_ok=False)
        with open(f"{folder}{file_name}", "w") as f:
            f.write(dumps(tasks, indent=1))
        f.close()

        print(f"0. {folder}{file_name} written.")

    def schedule(self, num_iteration, prices, model=None, solver=None, search=None, timeout=time_out):
        result = self.schedule_household(prices=prices,
                                         scheduling_method=self.scheduling_method,
                                         household=self.tasks,
                                         model=model, solver=solver, search=search, timeout=timeout)
        household_demand_profile = result[s_demand]
        weighted_penalty_household = result[s_penalty]
        self.household_tracker.update(num_record=num_iteration,
                                      demands=household_demand_profile,
                                      penalty=weighted_penalty_household)

        return household_demand_profile, weighted_penalty_household

    def schedule_household(self, prices, scheduling_method, household, num_intervals=no_intervals,
                           model=None, solver=None, search=None, timeout=time_out):

        prices = self.__convert_price(num_intervals, prices)

        # read tasks
        key = household[h_key]
        powers = household[h_powers]
        durations = household[h_durs]
        earliest_starts = household[h_ests]
        latest_ends = household[h_lfs]
        preferred_starts = household[h_psts]
        care_factors = household[h_cfs]
        max_care_factor = household[h_max_cf]
        precedents = [x[0] for x in list(household[h_precs].values())]
        successors = list(household[h_precs].keys())
        succ_delays = household[h_succ_delay]  # need to change this format when sending it to the solver
        no_precedents = household[h_no_precs]
        max_demand = household[h_demand_limit]
        inconvenience_cost_weight = household[h_incon_weight]

        # begin scheduling
        objective_values, big_value \
            = household_scheduling.preprocessing(powers=powers, durations=durations, max_demand=max_demand,
                                                 prices=prices,
                                                 preferred_starts=preferred_starts,
                                                 earliest_starts=earliest_starts,
                                                 latest_ends=latest_ends,
                                                 care_factors=care_factors,
                                                 inconvenience_cost_weight=inconvenience_cost_weight,
                                                 max_care_factor=max_care_factor, num_intervals=no_intervals)
        if "minizinc" in scheduling_method:
            model = file_cp_pre if model is None else model
            solver = solver_name if solver is None else solver
            search = f"int_search(actual_starts, {variable_selection}, {value_choice}, complete)" \
                if search is None else search
            succ_delays = [x[0] for x in list(household[h_succ_delay].values())]
            actual_starts, time_scheduling \
                = household_scheduling.minizinc_model(model_file=model, solver=solver, search=search,
                                                      objective_values=objective_values, powers=powers,
                                                      max_demand=max_demand,
                                                      durations=durations, earliest_starts=earliest_starts,
                                                      preferred_starts=preferred_starts,
                                                      latest_ends=latest_ends, successors=successors,
                                                      precedents=precedents,
                                                      no_precedents=no_precedents, succ_delays=succ_delays,
                                                      care_factors=care_factors,
                                                      prices=prices,
                                                      inconvenience_cost_weight=inconvenience_cost_weight,
                                                      num_intervals=num_intervals, timeout=timeout)
        else:
            actual_starts, time_scheduling \
                = household_scheduling.ogsa(objective_values=objective_values, big_value=big_value,
                                            powers=powers, durations=durations, preferred_starts=preferred_starts,
                                            latest_ends=latest_ends, max_demand=max_demand,
                                            successors=successors, precedents=precedents, succ_delays=succ_delays,
                                            randomness=False, num_intervals=num_intervals)

        household_demand_profile = [0] * num_intervals
        for p, ast, dur in zip(powers, actual_starts, durations):
            for t in range(ast, ast + dur):
                household_demand_profile[t % num_intervals] += p

        weighted_penalty_household \
            = inconvenience_cost_weight * sum([abs(pst - ast) * cf
                                               for pst, ast, cf in zip(preferred_starts, actual_starts, care_factors)])

        return {h_key: key, s_demand: household_demand_profile, s_starts: actual_starts,
                s_penalty: weighted_penalty_household, t_time: time_scheduling}

    def finalise_household(self, probability_distribution,
                           household_tracker_data=None, num_schedule=0):

        if household_tracker_data is None:
            household_tracker_data = self.household_tracker.data

        chosen_iter = choice(len(probability_distribution), size=1, p=probability_distribution)[0]
        chosen_demand_profile = household_tracker_data[s_demand][chosen_iter].copy()
        chosen_penalty = household_tracker_data[s_penalty][chosen_iter]

        if household_tracker_data is None:
            self.household_final.update(num_record=num_schedule,  demands=chosen_demand_profile, penalty=chosen_penalty)
        return chosen_demand_profile, chosen_penalty

    def __convert_price(self, num_intervals, prices):
        num_periods = len(prices)
        num_intervals_period = int(num_intervals / num_periods)
        if num_periods != num_intervals:
            prices = [p for p in prices for _ in range(num_intervals_period)]
        else:
            prices = [p for p in prices]

        return prices



from minizinc import *
import timeit
import random as r
from datetime import timedelta
from fw_ddsm.parameter import *


def preprocessing(
        powers, durations, max_demand, prices, preferred_starts, earliest_starts, latest_ends, care_factors,
        inconvenience_cost_weight, max_care_factor, num_intervals=no_intervals
):
    max_duration = max(durations)
    # this big cost and big cost * number_tasks need to be
    # smaller than the largest number that the solver can handle
    big_value = max_demand * max_duration * prices[-2] + \
                inconvenience_cost_weight * max_care_factor * num_intervals
    objective_value_matrix = []
    for power, pst, est, lft, dur, cf in zip(powers, preferred_starts, earliest_starts,
                                             latest_ends, durations, care_factors):
        objective_value_task = []
        for t in range(num_intervals):
            if est <= t <= lft - dur + 1:
                rc = abs(t - pst) * cf * inconvenience_cost_weight
                try:
                    rc += sum([prices[j % num_intervals] for j in range(t, t + dur)]) * power
                except IndexError:
                    print("Error: check the prices.")
            else:
                rc = big_value
            objective_value_task.append(int(rc))
        objective_value_matrix.append(objective_value_task)
    return objective_value_matrix, big_value


def minizinc_model(model_file, solver, search,
                   objective_values, powers, max_demand, durations,
                   earliest_starts, preferred_starts, latest_ends,
                   successors, precedents, no_precedents, succ_delays,
                   care_factors, prices, inconvenience_cost_weight,
                   num_intervals=no_intervals, timeout=time_out):
    # problem model
    model = Model(model_file)
    gecode = Solver.lookup(solver)
    model.add_string("solve ")
    if "gecode" in solver:
        model.add_string(":: {} ".format(search))
    model.add_string("minimize obj;")

    ins = Instance(gecode, model)
    num_tasks = len(powers)
    ins["num_intervals"] = num_intervals
    ins["num_tasks"] = num_tasks
    ins["durations"] = durations
    ins["demands"] = powers
    ins["num_precedences"] = no_precedents
    ins["predecessors"] = [p + 1 for p in precedents]
    ins["successors"] = [s + 1 for s in successors]
    ins["prec_delays"] = succ_delays
    ins["max_demand"] = max_demand

    if "ini" in model_type.lower():
        ins["prices"] = prices
        ins["preferred_starts"] = [ps + 1 for ps in preferred_starts]
        ins["earliest_starts"] = [es + 1 for es in earliest_starts]
        ins["latest_ends"] = [le + 1 for le in latest_ends]
        ins["care_factors"] = [cf * inconvenience_cost_weight for cf in care_factors]
    else:
        ins["run_costs"] = objective_values

    # solve problem model
    if timeout is None:
        result = ins.solve()
    else:
        result = ins.solve(timeout=timedelta(seconds=timeout))

    # process problem solution
    # obj = result.objective
    # print(result)
    solution = result.solution.actual_starts
    actual_starts = [int(a) - 1 for a in solution]
    # if "cp" in solver_type:
    #     actual_starts = [int(a) - 1 for a in solution]
    # else:  # "mip" in solver_type:
    #     actual_starts = [sum([i * int(v) for i, v in enumerate(row)]) for row in solution]
    time = result.statistics["time"].total_seconds()

    return actual_starts, time


def ogsa(objective_values, big_value, powers, durations, preferred_starts, latest_ends, max_demand,
         successors, precedents, succ_delays, randomness=True, num_intervals=no_intervals):
    start_time = timeit.default_timer()

    def retrieve_successors_or_precedents(list0, prec_or_succ_list1, succ_prec_list2):
        list_r = []
        for l in list0:
            if l in prec_or_succ_list1:
                succ_or_prec_indices = [i2 for i2, k in enumerate(prec_or_succ_list1) if k == l]
                succ_or_prec = [succ_prec_list2[i2] for i2 in succ_or_prec_indices]
                succ_succ_or_prec_prec \
                    = retrieve_successors_or_precedents(succ_or_prec, prec_or_succ_list1, succ_prec_list2)
                list_r.extend(succ_succ_or_prec_prec)
            else:
                list_r.append(int(l))
        return list_r

    def check_if_successors_or_precedents_exist(checked_task_id, prec_or_succ1, succ_or_prec2):
        succs_succs_or_precs_precs = []
        if checked_task_id in prec_or_succ1:
            indices = [i2 for i2, k in enumerate(prec_or_succ1) if k == checked_task_id]
            succs_or_precs = [succ_or_prec2[i2] for i2 in indices]
            succs_succs_or_precs_precs = retrieve_successors_or_precedents(succs_or_precs, prec_or_succ1,
                                                                           succ_or_prec2)
        return succs_succs_or_precs_precs

    actual_starts = []
    household_profile = [0] * num_intervals
    num_tasks = len(powers)
    for task_id in range(num_tasks):
        power = powers[task_id]
        duration = durations[task_id]
        task_costs = objective_values[task_id]

        # if i has successors
        tasks_successors = check_if_successors_or_precedents_exist(task_id, precedents, successors)
        earliest_suc_lstart_w_delay = 0
        earliest_suc_lstart = num_intervals - 1
        if bool(tasks_successors):
            suc_durations = [durations[i2] for i2 in tasks_successors]
            suc_lends = [latest_ends[i2] for i2 in tasks_successors]
            earliest_suc_lstart = min([lend - dur for lend, dur in zip(suc_lends, suc_durations)])
            # earliest_suc_lstart_w_delay = earliest_suc_lstart - succ_delay
            # suc_lstarts = [lend - dur + 1 for lend, dur in zip(suc_lends, suc_durations)]
            # earliest_suc_lstart = min(suc_lstarts)

        # if i has precedents
        tasks_precedents = check_if_successors_or_precedents_exist(task_id, successors, precedents)
        latest_pre_finish = 0
        latest_pre_finish_w_delay = num_intervals - 1
        if bool(tasks_precedents):
            prec_durations = [durations[i2] for i2 in tasks_precedents]
            prec_astarts = [actual_starts[i2] for i2 in tasks_precedents]
            succ_delay = succ_delays[task_id]
            latest_pre_finish = max([astart + dur - 1 for astart, dur in zip(prec_durations, prec_astarts)])
            latest_pre_finish_w_delay = latest_pre_finish + succ_delay[0]

        # search for all feasible intervals
        feasible_intervals = []
        for j in range(num_intervals):
            if task_costs[j] < big_value and earliest_suc_lstart_w_delay < j < earliest_suc_lstart - duration + 1 \
                    and latest_pre_finish < j < latest_pre_finish_w_delay:
                feasible_intervals.append(j)

        try:
            feasible_min_cost = min([task_costs[f] for f in feasible_intervals])
            cheapest_intervals = [f for f in feasible_intervals if task_costs[f] == feasible_min_cost]
            a_start = r.choice(cheapest_intervals) if randomness else cheapest_intervals[0]

            # check max demand constraint
            max_demand_starts = dict()
            temp_profile = household_profile[:]
            try:
                for d in range(a_start, a_start + duration):
                    temp_profile[d % num_intervals] += power
            except IndexError:
                print("error")
            temp_max_demand = max(temp_profile)
            while temp_max_demand > max_demand and len(feasible_intervals) > 1:

                max_demand_starts[a_start] = temp_max_demand
                feasible_intervals.remove(a_start)

                feasible_min_cost = min([objective_values[task_id][f] for f in feasible_intervals])
                feasible_min_cost_indices = [k for k, x in enumerate(objective_values[task_id]) if
                                             x == feasible_min_cost]
                # a_start = r.choice(feasible_min_cost_indices)
                a_start = feasible_min_cost_indices[0]

                temp_profile = household_profile[:]
                for d in range(a_start, a_start + duration):
                    temp_profile[d] += power
                temp_max_demand = max(temp_profile)

            if len(feasible_intervals) == 0 and not max_demand_starts:
                a_start = min(max_demand_starts, key=max_demand_starts.get)

        except ValueError:
            # print("No feasible intervals left for task", task_id)
            a_start = preferred_starts[task_id]

        actual_starts.append(a_start)

        # obj = 0

        # obj += objective_values[task_id][a_start]

    time_scheduling_ogsa = timeit.default_timer() - start_time
    # obj = round(obj, 2)

    return actual_starts, time_scheduling_ogsa



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fw_ddsm",
    version="0.2.38",
    author="Shan Dora He",
    author_email="dora.shan.he@gmail.com",
    description="Frank-Wolfe-based distributed demand scheduling method package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dorahee/FW-DDSM.git",
    packages=setuptools.find_packages(include=['fw_ddsm', 'fw_ddsm.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)